@@ -1,6 +1,6 @@
 
 // balance - a balancing tcp proxy
-// $Revision: 2.28 $
+// $Revision: 2.33 $
 //
 // Copyright (c) 2000,2001 by Thomas Obermair (obermair@acm.org) 
 // and Inlab Software GmbH (info@inlab.de), Gruenwald, Germany. 
@@ -22,8 +22,18 @@
 // 2.27:
 //   strcmp on NULL removed, thanks to Jay. D. Allen
 // 2.28: 
-//   bsent and breceived now unsigned to avoid nagative values,
+//   bsent and breceived now unsigned to avoid negative values,
 //   thanks to Anthony Baxter
+// 2.29:
+//   error in setaddress() fixed, thanks to Dirk Datzert
+// 2.30:
+//   fixing #includes for *bsd compability
+// 2.31: 
+// 2.32:
+//   redefied SIGCHLD handling to be compatible with FreeBSD 4.3,
+//   BSD/OS 4.2 and BSD/OS 4.0.1 
+// 2.33
+//   finally included SO_REUSEADDR
 //
 
 #include <stdio.h>
@@ -31,10 +41,8 @@
 #include <ctype.h>
 #include <errno.h>
 #include <netdb.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
 #include <signal.h>
-#include <malloc.h>
+#include <stdlib.h>
 
 #ifdef __FreeBSD__
 #define MYBSD 1
@@ -56,12 +64,16 @@
 
 #include <sys/socket.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <string.h>
 #include <sys/shm.h>
 
+#include <netinet/in.h>
+#include <arpa/inet.h>
+
 #define MAXTXSIZE 	(32*1024)
 #define FILENAMELEN 	1024
 #define SHMDIR 		"/tmp/"
@@ -71,8 +83,8 @@
 #define MAXINPUTLINE 	128	// max line in input mode 
 #define DEFAULTTIMEOUT  5	// default-timeout for non reachable hosts
 
-static char* rcsid="$Id: balance.c,v 2.28 2001/05/31 06:00:30 tommy Exp $";
-static char* revision="$Revision: 2.28 $";
+static char* rcsid="$Id: balance.c,v 2.33 2001/09/15 22:00:54 tommy Exp $";
+static char* revision="$Revision: 2.33 $";
 
 static char rendezvousfile[FILENAMELEN];
 static int  rendezvousfd;	
@@ -295,18 +307,25 @@ int setipaddress(struct in_addr *ipaddr, char* string) {
   }
 }
 
-int setaddress(struct in_addr *ipaddr, int *port, char* string, int default_port, int *maxc) {
-  char *host_string;
-  char *port_string;
-  char *maxc_string;
-
-  char *p, *q;
+void setaddress(struct in_addr *ipaddr, int *port, char* string, int default_port, int *maxc) {
+  char *host_string=NULL;
+  char *port_string=NULL;
+  char *maxc_string=NULL;
+  char *dup_string=NULL;
+  char *p=NULL;
+  char *q=NULL;
 
   struct hostent *hent;
 
-  host_string=string;
+  if((dup_string=strdup(string)) == NULL) {
+    fprintf(stderr,"strdup() failed\n");
+    exit(1);
+  }
 
-  if((p=index(string,':')) != NULL) {
+  host_string=dup_string;
+  p=index(dup_string,':');
+
+  if(p != NULL) {
     *p='\000';
     port_string=p+1;
     if((q=index(port_string,':')) != NULL) {
@@ -317,6 +336,7 @@ int setaddress(struct in_addr *ipaddr, int *port, char* string, int default_port
     }
   } else {
     port_string="";
+    maxc_string="";
   }
 
   // fix for RedHat 7.0/7.1 choke on strcmp with NULL
@@ -326,10 +346,10 @@ int setaddress(struct in_addr *ipaddr, int *port, char* string, int default_port
   if(maxc_string != NULL && !strcmp(maxc_string,"")) 
     maxc_string=NULL; 
   
-  hent = gethostbyname(string);
+  hent = gethostbyname(dup_string);
   if(hent == NULL) {
-    if((ipaddr->s_addr = inet_addr(string))== -1) {
-      fprintf(stderr,"unknown or invalid address [%s]\n",string);
+    if((ipaddr->s_addr = inet_addr(dup_string))== -1) {
+      fprintf(stderr,"unknown or invalid address [%s]\n",dup_string);
       exit(1);
     }
   } else {
@@ -345,6 +365,7 @@ int setaddress(struct in_addr *ipaddr, int *port, char* string, int default_port
   if(maxc_string != NULL) {
     *maxc=atoi(maxc_string);
   }
+  free(dup_string);
 }
 
 int setaddress_noexitonerror(struct in_addr *ipaddr, int *port, char* string, int default_port) {
@@ -509,6 +530,12 @@ stream2(int clientfd, int serverfd, int groupindex, int channelindex) {
 void alarm_handler() {
 }
 
+void chld_handler() {
+  int status;
+  while(waitpid(-1, &status, WNOHANG) > 0);
+  (void) signal(SIGCHLD, chld_handler);
+}
+
 // a channel in a group is selected and we try to establish a connection 
 
 void *stream(int arg, int groupindex, int index) {
@@ -704,8 +731,8 @@ hintergrund() {
   setpgid(getpid(),0);
 #else
   setpgrp();
-  signal(SIGCHLD,SIG_IGN);
 #endif
+  signal(SIGCHLD,chld_handler);
 }
 
 COMMON* makecommon(int argc, char** argv, int source_port) {
@@ -971,7 +998,7 @@ shell() {
         b_unlock();
 
       } else if(mycmp(command,"version")) {
-	printf("  $Id: balance.c,v 2.28 2001/05/31 06:00:30 tommy Exp $\n");
+	printf("  $Id: balance.c,v 2.33 2001/09/15 22:00:54 tommy Exp $\n");
         printf("  MAXGROUPS=%d\n", MAXGROUPS);
         printf("  MAXCHANNELS=%d\n", MAXCHANNELS);
       } else {
@@ -1128,7 +1155,7 @@ char	*argv[];
 
   // --
 
-  signal(SIGCHLD, SIG_IGN);
+  signal(SIGCHLD, chld_handler);
 
   if(!foreground) {
     hintergrund();
@@ -1154,6 +1181,15 @@ char	*argv[];
     err_dump("can't open stream socket");
   }
 
+  { 
+    int sockopton = 1;
+    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, 
+         (char *)&sockopton, sizeof(int) ) == -1 ) {
+      perror("setsockopt(SO_REUSEADDR)");
+      err_dump("can't so_reuseaddr");
+    }
+  }
+
   if(debugflag) {
     fprintf(stderr,"bindhost_adress=[%s]\n",bindhost_address);
   }
@@ -1168,16 +1204,18 @@ char	*argv[];
 
   listen(sockfd, SOMAXCONN);
 
+/*
 #ifdef MYBSD
   act.sa_handler = sig_chld;
   sigemptyset(&act.sa_mask);
-  act.sa_flags = SA_NOCLDSTOP;
+  act.sa_flags = SA_NOCLDSTOP|SA_NOCLDWAIT;
   if (sigaction(SIGCHLD, &act, NULL) < 0) {
     fprintf(stderr, "sigaction failed\n");
     return 1;
   }
 #else
 #endif
+*/
 
   for (;;) {
     int index;
